/*
 * ntpclient.c - NTP client
 *
 * Copyright 1997, 1999, 2000, 2003, 2006, 2007  Larry Doolittle  <larry@doolittle.boa.org>
 * Last hack: December 30, 2007
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License (Version 2,
 *  June 1991) as published by the Free Software Foundation.  At the
 *  time of writing, that license was published by the FSF with the URL
 *  http://www.gnu.org/copyleft/gpl.html, and is incorporated herein by
 *  reference.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  Possible future improvements:
 *      - Write more documentation  :-(
 *      - Support leap second processing
 *      - Support IPv6
 *      - Support multiple (interleaved) servers
 *
 *  Compile with -DPRECISION_SIOCGSTAMP if your machine really has it.
 *  There are patches floating around to add this to Linux, but
 *  usually you only get an answer to the nearest jiffy.
 *  Hint for Linux hacker wannabes: look at the usage of get_fast_time()
 *  in net/core/dev.c, and its definition in kernel/time.c .
 *
 *  If the compile gives you any flak, check below in the section
 *  labelled "XXX fixme - non-automatic build configuration".
 *
 *  If porting ntpclient into new environment check the size of u32
 *  being 32: just define once CHECK_SIZEOF_INT_BEING_32_BITS
 *
 * There are define macros to control the logging of the code:
 * -DENABLE_STDERR, -DENABLE_DEBUG and -DENABLE_DEBUG2SYSLOG
 *
 */

#define _POSIX_C_SOURCE 199309

#ifdef USE_OBSOLETE_GETTIMEOFDAY
#define _BSD_SOURCE
#endif

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netdb.h>     /* gethostbyname */
#include <arpa/inet.h>
#include <time.h>
#include <unistd.h>
#include <errno.h>
#include <signal.h>
#ifdef PRECISION_SIOCGSTAMP
#include <sys/ioctl.h>
#endif
#ifdef USE_OBSOLETE_GETTIMEOFDAY
#include <sys/time.h>
#endif

#include <syslog.h>

#include "ntpclient.h"

/* Default to the RFC-4330 specified value */
#ifndef MIN_INTERVAL
#define MIN_INTERVAL 15
#endif

#ifdef ENABLE_DEBUG
#define DEBUG_OPTION "d"
int debug=0;
#else
#define DEBUG_OPTION
#endif

/* define DEBUG Possibilities:
    ENABLE_DEBUG2SYSLOG                 - debug info goes into syslog
    int debug2syslog                    - in case of ClockService having DEBUG trace level, 
                                          the ntpclient is called with a parameter '-L'.
                                          The debug info is depending on it
    ENABLE_DEBUG_IPADDR                 - In setup_receive_generic() the ip addresses is debugged:
                                          if defined then the ip address is sysouted else not                                 
    DEFINE_IN_NEED__ENABLE_DEBUG2SYSLOG - special case, not used normaly, but can be inserted
*/

/* MRH93817 needed more syslog possibilities */
#ifdef ENABLE_DEBUG2SYSLOG
	#define DEBUG2SYSLOG_OPTION "L"
	int debug2syslog=0;
#else
	#define DEBUG2SYSLOG_OPTION
#endif

#ifdef ENABLE_REPLAY
#define  REPLAY_OPTION   "r"
#else
#define  REPLAY_OPTION
#endif

extern char *optarg;  /* according to man 2 getopt */

#include <stdint.h>
typedef uint32_t u32;  /* universal for C99 */
/* typedef u_int32_t u32;   older Linux installs? */

/* XXX fixme - non-automatic build configuration */
#ifdef __linux__
#include <sys/utsname.h>
#include <sys/time.h>
#include <sys/timex.h>
#else
extern struct hostent *gethostbyname(const char *name);
extern int h_errno;
#endif
/* end configuration for host systems */

#define JAN_1970        0x83aa7e80      /* 2208988800 1970 - 1900 in seconds */
#define NTP_PORT "123"

/* How to multiply by 4294.967296 quickly (and not quite exactly)
 * without using floating point or greater than 32-bit integers.
 * If you want to fix the last 12 microseconds of error, add in
 * (2911*(x))>>28)
 */
#define NTPFRAC(x) ( 4294*(x) + ( (1981*(x))>>11 ) )

/* The reverse of the above, needed if we want to set our microsecond
 * clock (via clock_settime) based on the incoming time in NTP format.
 * Basically exact.
 */
#define USEC(x) ( ( (x) >> 12 ) - 759 * ( ( ( (x) >> 10 ) + 32768 ) >> 16 ) )

/* Converts NTP delay and dispersion, apparently in seconds scaled
 * by 65536, to microseconds.  RFC-1305 states this time is in seconds,
 * doesn't mention the scaling.
 * Should somehow be the same as 1000000 * x / 65536
 */
#define sec2u(x) ( (x) * 15.2587890625 )

struct ntptime
{
  unsigned int coarse;
  unsigned int fine;
};

struct ntp_control
{
  u32 time_of_send[2];
  int live;
  int set_clock;   /* non-zero presumably needs root privs */
  int probe_count;
  int cycle_time;
  int timeout;
  int goodness;
  int cross_check;
  char serv_addr[4];
};

short signalexit=0;      //value of exit signal MRI14137
/* prototypes for some local routines */
static void send_packet(int usd, u32 time_sent[2]);
static int rfc1305print(u32 *data, struct ntptime *arrival, struct ntp_control *ntpc, int *error);
void signal_handler(int); // function for catch the SIG_TERM signal according to MRI14137
/* static void udp_handle(int usd, char *data, int data_len, struct sockaddr *sa_source, int sa_len); */

static const char* ntpclient_pipe = "/tmp/gpal/ipc/ntpclient";

static int get_current_freq(void)
{
  /* OS dependent routine to get the current value of clock frequency.
   */
#ifdef __linux__
  struct timex txc;
  txc.modes=0;
  if (__adjtimex(&txc) < 0)
  {
#ifdef ENABLE_STDERR
    perror("adjtimex");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.get_current_freq: adjtimex error: %s", strerror(errno));
#endif
  }
  return txc.freq;
#else
  return 0;
#endif
}


static int set_freq(int new_freq)
{
  /* OS dependent routine to set a new value of clock frequency.
   */
#ifdef __linux__
  struct timex txc;
  txc.modes = ADJ_FREQUENCY;
  txc.freq = new_freq;
  if (__adjtimex(&txc) < 0)
  {
#ifdef ENABLE_STDERR
    perror("adjtimex");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.set_freq: adjtimex error: %s", strerror(errno));
#endif
  }
  return txc.freq;
#else /* __linux__ */
  return 0;
#endif /* __linux__ */
}


static void set_time(struct ntptime *new)
{
  FILE *fp;
  char buffer[128];
  int len;

#ifndef USE_OBSOLETE_GETTIMEOFDAY
  /* POSIX 1003.1-2001 way to set the system clock
   */
  struct timespec tv_set;
  /* it would be even better to subtract half the slop */
  tv_set.tv_sec  = new->coarse - JAN_1970;
  /* divide xmttime.fine by 4294.967296 */
  tv_set.tv_nsec = USEC(new->fine)*1000;
  if (clock_settime(CLOCK_REALTIME, &tv_set)<0)
  {
#ifdef ENABLE_STDERR
    perror("clock_settime");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.set_time: clock_settime error: %s. Stopped.", strerror(errno));
#endif
    exit(1);
  }
#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("set time to %lu.%.9lu\n", tv_set.tv_sec, tv_set.tv_nsec);
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_ERR, "NTPclient.set_time: set time to %lu.%.9lu", tv_set.tv_sec, tv_set.tv_nsec);
  }
#endif
#else /* USE_OBSOLETE_GETTIMEOFDAY */
  /* Traditional Linux way to set the system clock
   */
  struct timeval tv_set;

  /* it would be even better to subtract half the slop */
  tv_set.tv_sec  = new->coarse - JAN_1970;
  /* divide xmttime.fine by 4294.967296 */
  tv_set.tv_usec = USEC(new->fine);
  if (settimeofday(&tv_set,NULL)<0)
  {
#ifdef ENABLE_STDERR
    perror("settimeofday");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.set_time: settimeofday error: %s. Stopped.", strerror(errno));
#endif
    exit(1);
  }
#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("set time to %lu.%.6lu\n", tv_set.tv_sec, tv_set.tv_usec);
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_ERR, "NTPclient.set_time: set time to %lu.%.6lu", tv_set.tv_sec, tv_set.tv_usec);
  }
#endif
#endif /* USE_OBSOLETE_GETTIMEOFDAY */

  syslog(LOG_DEBUG, "%s %s", "NTPclient.set_time pipes time: ", ctime(&tv_set.tv_sec));

  fp = fopen(ntpclient_pipe, "w");
  if (fp)
  {
    len = snprintf(buffer, sizeof(buffer)-1, "%lu", tv_set.tv_sec);
    fflush(stdout);
    fwrite(buffer, len, 1, fp);
    fclose(fp);
  }
}


static void ntpc_gettime(u32 *time_coarse, u32 *time_fine)
{
#ifndef USE_OBSOLETE_GETTIMEOFDAY
  /* POSIX 1003.1-2001 way to get the system time
   */
  struct timespec now;
  clock_gettime(CLOCK_REALTIME, &now);
  *time_coarse = now.tv_sec + JAN_1970;
  *time_fine   = NTPFRAC(now.tv_nsec/1000);
#else
  /* Traditional Linux way to get the system time
   */
  struct timeval now;
  gettimeofday(&now, NULL);
  *time_coarse = now.tv_sec + JAN_1970;
  *time_fine   = NTPFRAC(now.tv_usec);
#endif /* USE_OBSOLETE_GETTIMEOFDAY */
}


static void send_packet(int usd, u32 time_sent[2])
{
  u32 data[12];
#define LI 0
#define VN 3
#define MODE 3
#define STRATUM 0
#define POLL 4
#define PREC -6

#if defined(ENABLE_DEBUG) && defined(ENABLE_STDERR)
  if (debug)
  {
    fprintf(stderr,"Sending ...\n");
  }
#endif

  /* ----------------------------------------------------------------------------------------- */
  /* Removed checking of sizeof(data) being 48 in real-time (at every call of send_packet()).  */
  /* It is sufficient to check 'data' once, if ntpclient is newly ported -> use define:        */
  /* #define CHECK_SIZEOF_INT_BEING_32_BITS                                                    */
  /* ----------------------------------------------------------------------------------------- */
#ifdef  CHECK_SIZEOF_INT_BEING_32_BITS
  if (sizeof data != 48)
  {
#ifdef ENABLE_STDERR
    fprintf(stderr,"size error\n");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.send_packet: size error: %d != 48", sizeof(data));
#endif
    return;
  }
  /* ----------------------------------------------------------------------------------------- */
#endif /* CHECK_SIZEOF_INT_BEING_32_BITS                                                     */
  /* ----------------------------------------------------------------------------------------- */

  memset(data,0,sizeof data);
  data[0] = htonl (
              ( LI << 30 ) | ( VN << 27 ) | ( MODE << 24 ) |
              ( STRATUM << 16) | ( POLL << 8 ) | ( PREC & 0xff ) );
  data[1] = htonl(1<<16);  /* Root Delay (seconds) */
  data[2] = htonl(1<<16);  /* Root Dispersion (seconds) */
  ntpc_gettime(time_sent, time_sent+1);
  data[10] = htonl(time_sent[0]); /* Transmit Timestamp coarse */
  data[11] = htonl(time_sent[1]); /* Transmit Timestamp fine   */
  send(usd,data,48,0);
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_DEBUG, "NTPclient.send_packet - sent");
  }
#endif
}


static void get_packet_timestamp(int usd, struct ntptime *udp_arrival_ntp)
{
#ifdef PRECISION_SIOCGSTAMP
  /* XXX broken */
  struct timeval udp_arrival;
  if ( ioctl(usd, SIOCGSTAMP, &udp_arrival) < 0 )
  {
#ifdef ENABLE_STDERR
    perror("ioctl-SIOCGSTAMP");
#endif
    gettimeofday(&udp_arrival,NULL);
  }
  udp_arrival_ntp->coarse = udp_arrival.tv_sec + JAN_1970;
  udp_arrival_ntp->fine   = NTPFRAC(udp_arrival.tv_usec);
#else /* PRECISION_SIOCGSTAMP */
  (void) usd;  /* not used */
  ntpc_gettime(&udp_arrival_ntp->coarse, &udp_arrival_ntp->fine);
#endif /* PRECISION_SIOCGSTAMP */
}


/**
 * check_source() check nothing but logs the data_len and net address.
 * it is an unfortunate naming.
 * returns 0 everytime.
 */
static int check_source(int data_len, struct sockaddr_storage *sa_source, unsigned int sa_len, struct ntp_control *ntpc)
{
  char tmp[80]; /* 15 x 'ffff:' + 'ffff\0'-- len=80 */
  struct sockaddr_in *sa_in=(struct sockaddr_in *)sa_source;
  struct sockaddr_in6 *sa_in6=(struct sockaddr_in6 *)sa_source;
  (void) sa_len;  /* not used */

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("packet of length %d received\n",data_len);
    if (sa_source->ss_family==AF_INET)
    {
      printf("Source: AF_INET Port '%d' host '%s'\n",
             ntohs(sa_in->sin_port),inet_ntoa(sa_in->sin_addr));
    }
    else if (sa_source->ss_family==AF_INET6)
    {
      printf("Source: AF_INET6 Port '%d' host '%s'\n",
             ntohs(sa_in6->sin6_port),inet_ntop(AF_INET6, (const void *)sa_in6->sin6_addr.s6_addr, (char*)tmp, sizeof(tmp)));
    }
    else
    {
      printf("Source sa_source->ss_family: '%d' [!2(AF_INET) !10(AF_INET6)]\n",
             sa_source->ss_family);
    }
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_DEBUG, "NTPclient.check_source: packet received (l=%d)", data_len);
    if (sa_source->ss_family==AF_INET)
    {
      syslog(LOG_DEBUG, "NTPclient.check_source: AF_INET Port '%d' host '%s'",
             ntohs(sa_in->sin_port),inet_ntoa(sa_in->sin_addr));
    }
    else if (sa_source->ss_family==AF_INET6)
    {
      syslog(LOG_DEBUG, "NTPclient.check_source: AF_INET6 Port '%d' host '%s'",
             ntohs(sa_in6->sin6_port),inet_ntop(AF_INET6, (const void *)sa_in6->sin6_addr.s6_addr, (char*)tmp, sizeof(tmp)));
    }
    else
    {
      syslog(LOG_ERR, "NTPclient.check_source: ss_family: '%d' (AF_INET=2, AF_INET6=10)", sa_source->ss_family);
    }
  }
#endif
  /* we could check that the source is the server we expect, but
   * Denis Vlasenko recommends against it: multihomed hosts get it
   * wrong too often. */
#if 0
  if (memcmp(ntpc->serv_addr, &(sa_in->sin_addr), 4)!=0)
  {
    return 1;  /* fault */
  }
#else
  (void) ntpc; /* not used */
#endif
#if 0
  if (NTP_PORT != ntohs(sa_in->sin_port))
  {
    return 1;  /* fault */
  }
#endif
  return 0;
}


static double ntpdiff( struct ntptime *start, struct ntptime *stop)
{
  int a;
  unsigned int b;
  a = stop->coarse - start->coarse;
  if (stop->fine >= start->fine)
  {
    b = stop->fine - start->fine;
  }
  else
  {
    b = start->fine - stop->fine;
    b = ~b;
    a -= 1;
  }

  return a*1.e6 + b * (1.e6/4294967296.0);
}


/* Does more than print, so this name is bogus.
 * It also makes time adjustments, both sudden (-s)
 * and phase-locking (-l).
 * sets *error to the number of microseconds uncertainty in answer
 * calls "set_time()" depending on the sanity checks
 * returns 0 normally, 1 if the message fails sanity checks
 */
static int rfc1305print(u32 *data, struct ntptime *arrival, struct ntp_control *ntpc, int *error)
{
  /* straight out of RFC-1305 Appendix A */
  int li, vn, mode, stratum, poll, prec;
  int delay, disp, refid;
  struct ntptime reftime, orgtime, rectime, xmttime;
  double el_time,st_time,skew1,skew2;
  int freq;

#define Data(i) ntohl(((u32 *)data)[i])
  li      = Data(0) >> 30 & 0x03;
  vn      = Data(0) >> 27 & 0x07;
  mode    = Data(0) >> 24 & 0x07;
  stratum = Data(0) >> 16 & 0xff;
  poll    = Data(0) >>  8 & 0xff;
  prec    = Data(0)       & 0xff;
  if (prec & 0x80) prec|=0xffffff00;
  delay   = Data(1);
  disp    = Data(2);
  refid   = Data(3);
  reftime.coarse = Data(4);
  reftime.fine   = Data(5);
  orgtime.coarse = Data(6);
  orgtime.fine   = Data(7);
  rectime.coarse = Data(8);
  rectime.fine   = Data(9);
  xmttime.coarse = Data(10);
  xmttime.fine   = Data(11);
#undef Data

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("LI=%d  VN=%d  Mode=%d  Stratum=%d  Poll=%d  Precision=%d\n",
           li, vn, mode, stratum, poll, prec);
    printf("Delay=%.1f  Dispersion=%.1f  Refid=%u.%u.%u.%u\n",
           sec2u(delay),sec2u(disp),
           refid>>24&0xff, refid>>16&0xff, refid>>8&0xff, refid&0xff);
    printf("Reference %u.%.6u, (sent)    %u.%.6u\n", reftime.coarse, USEC(reftime.fine), ntpc->time_of_send[0], USEC(ntpc->time_of_send[1]));
    printf("Originate %u.%.6u, Receive   %u.%.6u\n", orgtime.coarse, USEC(orgtime.fine), rectime.coarse, USEC(rectime.fine));
    printf("Transmit  %u.%.6u, Our recv  %u.%.6u\n", xmttime.coarse, USEC(xmttime.fine), arrival->coarse, USEC(arrival->fine));
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: LI=%d  VN=%d  Mode=%d  Stratum=%d  Poll=%d  Precision=%d",
           li, vn, mode, stratum, poll, prec);
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: Delay=%.1f  Dispersion=%.1f  Refid=%u.%u.%u.%u",
           sec2u(delay),sec2u(disp),refid>>24&0xff, refid>>16&0xff, refid>>8&0xff, refid&0xff);
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: Reference %u.%.6u, (sent)    %u.%.6u",
           reftime.coarse, USEC(reftime.fine), ntpc->time_of_send[0], USEC(ntpc->time_of_send[1]));
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: Originate %u.%.6u, Receive   %u.%.6u",
           orgtime.coarse, USEC(orgtime.fine), rectime.coarse, USEC(rectime.fine));
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: Transmit  %u.%.6u, Our recv  %u.%.6u",
           xmttime.coarse, USEC(xmttime.fine), arrival->coarse, USEC(arrival->fine));
  }
#endif
  el_time=ntpdiff(&orgtime,arrival);   /* elapsed */
  st_time=ntpdiff(&rectime,&xmttime);  /* stall */
  skew1=ntpdiff(&orgtime,&rectime);
  skew2=ntpdiff(&xmttime,arrival);
  freq=get_current_freq();
#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("Total elapsed: %9.2f\n"
           "Server stall:  %9.2f\n"
           "Slop:          %9.2f\n",
           el_time, st_time, el_time-st_time);
    printf("Skew:          %9.2f\n"
           "Frequency:     %9d\n",
           (skew1-skew2)/2, freq);
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: [Total elapsed: %9.2f][Server stall: %9.2f][Slop: %9.2f][Skew: %9.2f][Frequency: %9d]",
           el_time, st_time, el_time-st_time,(skew1-skew2)/2, freq);
  }
#endif

  /* error checking, see RFC-4330 section 5 */
  static const char * errLI = "'LI==3'";
  static const char * errVN = "'VN<3'";
  static const char * errMODE = "'MODE!=3'";
  static const char * errORG = "'ORG!=sent'";
  static const char * errXMT = "'XMT==0'";
  static const char * errDELAY = "'abs(DELAY)>65536'";
  static const char * errDISP = "'abs(DISP)>65536'";
  static const char * errSTRATUM = "'STRATUM==0'";
  static const char * errAnd = " and ";
#define FAIL_STR_MAXLEN (256)
  char fail_str[FAIL_STR_MAXLEN];
  int fail_marker=0;

  /* Here starts the sanity check ...  */
#if defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG)
#define FAIL(x) { \
  if (fail_marker) \
  { \
   if (FAIL_STR_MAXLEN - strlen(fail_str) - strlen(errAnd) > 1 ) strcat(fail_str,errAnd); \
  } \
  fail_marker++; \
  if (FAIL_STR_MAXLEN - strlen(fail_str) - strlen(x) > 1 ) strcat(fail_str,x); \
 }
#else  /* defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG) */
#define FAIL(x) \
 { \
  fail_marker++; \
 }
#endif /* defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG) */
  if (ntpc->cross_check)
  {
    fail_str[0] = 0;
    if (li == 3) FAIL(errLI);  /* unsynchronized */
    if (vn < 3) FAIL(errVN);   /* RFC-4330 documents SNTP v4, but we interoperate with NTP v3 */
    if (mode != 4) FAIL(errMODE);
    if (orgtime.coarse != ntpc->time_of_send[0] ||
        orgtime.fine   != ntpc->time_of_send[1] ) FAIL(errORG);
    if (xmttime.coarse == 0 && xmttime.fine == 0) FAIL(errXMT);
    if (delay > 65536 || delay < -65536) FAIL(errDELAY);
    if (disp  > 600000 || disp  < -600000) FAIL(errDISP); //Allow 10 sec root dispersion I01635 
    if (stratum == 0) FAIL(errSTRATUM);  /* kiss o' death */
  }
#undef FAIL

#if defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG)
  if (fail_marker)
  { /* error */
#ifdef ENABLE_DEBUG
    if (debug)
    {
      printf("%d %.5d.%.3d  packet rejected because %s\n",
             arrival->coarse/86400, arrival->coarse%86400,
             arrival->fine/4294967, fail_str);
    }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_DEBUG, "NTPclient.rfc1305print: %d %.5d.%.3d  packet rejected because %s",
           arrival->coarse/86400, arrival->coarse%86400,
           arrival->fine/4294967, fail_str);
#endif
  }
#endif /* defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG) */

  if (!fail_marker)
  { /* correct packet produced and sets time */
#if defined(ENABLE_DEBUG) || defined(ENABLE_DEBUG2SYSLOG)

#ifdef ENABLE_DEBUG
    if (debug)
    {
      printf("NTPclient.rfc1305print(): no failmarker set\n");
    }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_DEBUG, "NTPclient.rfc1305print:no failmarker set");
#endif
#endif
    /* XXX should I do this if debug flag is set? */
    if (ntpc->set_clock)   /* you'd better be root, or ntpclient will crash! */
    {
      set_time(&xmttime);
    }

    /* Not the ideal order for printing, but we want to be sure
     * to do all the time-sensitive thinking (and time setting)
     * before we start the output, especially fflush() (which
     * could be slow).  Of course, if debug is turned on, speed
     * has gone down the drain anyway. */
    if (ntpc->live)
    {
      int new_freq;
      new_freq = contemplate_data(arrival->coarse, (skew1-skew2)/2,
                                  el_time+sec2u(disp), freq);
#ifdef ENABLE_DEBUG
      if (!debug && new_freq != freq)
      {
        set_freq(new_freq);
      }
#else
      if (new_freq != freq)
      {
        set_freq(new_freq);
      }
#endif /* ENABLE_DEBUG */
    }

#ifdef ENABLE_DEBUG
    if (debug)
    {
      printf(" day   second     elapsed    stall     skew  dispersion  freq\n");
      printf("%d %.5d.%.3d  %8.1f %8.1f  %8.1f %8.1f %9d\n",
             arrival->coarse/86400, arrival->coarse%86400,
             arrival->fine/4294967, el_time, st_time,
             (skew1-skew2)/2, sec2u(disp), freq);
      fflush(stdout);
    }
#endif
#ifdef DEFINE_IN_NEED__ENABLE_DEBUG2SYSLOG
    if (debug2syslog)
    {
      syslog(LOG_DEBUG, "NTPclient.rfc1305print: day   second     elapsed    stall     skew  dispersion  freq");
      syslog(LOG_DEBUG, "NTPclient.rfc1305print: %d %.5d.%.3d  %8.1f %8.1f  %8.1f %8.1f %9d",
             arrival->coarse/86400, arrival->coarse%86400,
             arrival->fine/4294967, el_time, st_time,
             (skew1-skew2)/2, sec2u(disp), freq);
    }
#endif
    *error = el_time-st_time;
  }

  return (!fail_marker) ? 0 : 1 ;
}


#if 0
/* original IPv4 only implementation */
static void stuff_net_addr(struct in_addr *p, char *hostname)
{
  struct hostent *ntpserver;
  ntpserver=gethostbyname(hostname);
  if (ntpserver == NULL)
  {
    fprintf(stderr,"Error %d looking up hostname %s\n", h_errno,hostname)
    exit(1);
  }
  if (ntpserver->h_length != 4)
  {
    /* IPv4 only, until I get a chance to test IPv6 */
    fprintf(stderr,"oops %d\n",ntpserver->h_length);
    exit(1);
  }
  memcpy(&(p->s_addr),ntpserver->h_addr_list[0],4);
}
#endif


/* added support of IPv6 */
int stuff_net_addr_generic(const char *hostname,
                           const char *service,
                           int         family,
                           int         socktype)
{
  struct addrinfo hints, *res, *ressave;
  int error, sockfd, binderr;

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("stuff_net_addr_generic\n");
  }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
  if (debug2syslog)
  {
    syslog(LOG_DEBUG, "NTPclient.stuff_net_addr_generic(hostname=\"%s\",service=\"%s\",family=(%d),socktype=(%d))", (hostname) ? hostname : "NULL",service,family,socktype);
  }
#endif

  memset(&hints, 0, sizeof(struct addrinfo));

  hints.ai_flags    = AI_PASSIVE;
  hints.ai_family   = family;
  hints.ai_socktype = socktype;

  error = getaddrinfo(hostname, service, &hints, &res);

  if (error != 0)
  {
#ifdef ENABLE_STDERR
    fprintf(stderr,
            "getaddrinfo error: [%s]\n", gai_strerror(error));
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.getaddrinfo error: [%s]. Stopped.", gai_strerror(error));
#endif
    return -1;
  }

  ressave = res;

  sockfd=-1;
  while (res)
  {
    errno=0;
    sockfd = socket(res->ai_family,
                    res->ai_socktype,
                    res->ai_protocol);

#ifdef ENABLE_DEBUG2SYSLOG
    if (debug2syslog)
    {
        syslog(LOG_DEBUG,"NTPclient.stuff_net_addr_generic: socket(family=(%d), socktype=(%d), protocol=(%d)) => sockfd=(%d) errno=(%d)", res->ai_family,res->ai_socktype,res->ai_protocol,sockfd,errno);
    }
#endif
    if (sockfd >= 0)
    {
      errno=0;
      binderr=bind(sockfd, res->ai_addr, res->ai_addrlen);
#ifdef ENABLE_DEBUG2SYSLOG
      if (debug2syslog)
      {
        syslog(LOG_DEBUG, "NTPclient.stuff_net_addr_generic: bind(sockfd=(%d),addrlen=(%d)) => error=%d, errno=(%d)", sockfd, res->ai_addrlen, binderr, errno);
      }
#endif
      if (binderr == 0)
      {
#ifdef ENABLE_DEBUG
        if (debug)
        {
          printf("NTPclient.stuff_net_addr_generic(): bind successful on socket %d\n", sockfd);
        }
#endif
#ifdef DEFINE_IN_NEED__ENABLE_DEBUG2SYSLOG
        if (debug2syslog)
        {
          syslog(LOG_DEBUG, "NTPclient.stuff_net_addr_generic: bind successful on socket %d", sockfd);
        }
#endif
        break;
      }
      close(sockfd);
      sockfd=-1;
    }
    res=res->ai_next;
  }

  freeaddrinfo(ressave);

#ifdef ENABLE_DEBUG2SYSLOG
  if (sockfd==-1)
  {
    syslog(LOG_ERR, "NTPclient.stuff_net_addr_generic: could NOT bind on socket '%d'. Stopped.", sockfd);
  }
#endif

  return sockfd;
}


#if IPv4_ONLY
/* original IPv4 only implementation */
static void setup_receive(int usd, unsigned int interface, short port)
{
  struct sockaddr_in sa_rcvr;
  memset(&sa_rcvr,0,sizeof sa_rcvr);
  sa_rcvr.sin_family=AF_INET;
  sa_rcvr.sin_addr.s_addr=htonl(interface);
  sa_rcvr.sin_port=htons(port);
  if (bind(usd,(struct sockaddr *) &sa_rcvr,sizeof sa_rcvr) == -1)
  {
#ifdef ENABLE_STDERR
    perror("bind");
    fprintf(stderr,"could not bind to udp port %d\n",port);
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    syslog(LOG_ERR, "NTPclient.setup_receive: could not bind to udp port '%d': %s. Stopped.", strerror(errno));
#endif
    exit(1);
  }
  /* listen(usd,3); this isn't TCP; thanks Alexander! */
}
#endif /* IPv4_ONLY */


/* added support of IPv6 */
int setup_receive_generic(char *hostname,
                          const char *service,
                          int         family,
                          int         socktype,
                          int         fd)
{
  struct addrinfo hints, *res, *ressave;
  int error;
  // MR H95366 new return value will be checked
  int conErr=0;
  int delay_in_sec, counter;
#define BACK_OFF_COUNTER_MAX (8)

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("setup_receive_generic entered\n");
  }
#endif

  delay_in_sec = 1;
  counter  = 0;      // 1 - 2 - 4 - 8 - 16 - 32 - 64 - 128 sec pauses //
  while (1)
  {
    memset(&hints, 0, sizeof(struct addrinfo));

    hints.ai_family   = family;
    hints.ai_socktype = socktype;

    error = getaddrinfo(hostname, service, &hints, &res);

    if (error != 0)
    {
#ifdef ENABLE_STDERR
      fprintf(stderr,
              "getaddrinfo error: [%s]\n", gai_strerror(error));
#endif
#ifdef ENABLE_DEBUG2SYSLOG
      syslog(LOG_ERR, "NTPclient.setup_receive_generic: getaddrinfo error: '%s'", gai_strerror(error));
#endif
      return -1;
    }

    ressave = res;
    while (res)
    {
      errno = 0;
      conErr = connect(fd, res->ai_addr, res->ai_addrlen);

#ifdef ENABLE_DEBUG2SYSLOG
      if (debug2syslog)
      {
#ifdef ENABLE_DEBUG_IPADDR
        static const char ascII[]="0123456789ABCDEF";
        char adrStr[256];
        unsigned int   i,j;
        for (i=j=0; ((i < res->ai_addrlen) && (j <= 252)); i++)		// worst case [250]=xxyy \0
        {
          adrStr[j++] = ascII[(res->ai_addr->sa_data[i]>>4) & 15];
          adrStr[j++] = ascII[(res->ai_addr->sa_data[i]   ) & 15];
          if ((i%2) && (i  <  (res->ai_addrlen-1)))
            adrStr[j++] = ' ';
        }
        adrStr[j] = 0;
        syslog(LOG_ERR, "NTPclient.setup_receive_generic: connect(adr:[fam:%d]:\"%s\", len:%d) ==> result=%d, errno=%d", res->ai_addr->sa_family, adrStr, res->ai_addrlen, conErr, errno);
#else
        syslog(LOG_ERR, "NTPclient.setup_receive_generic: connect(adr:[fam:%d], len:%d) ==> result=%d, errno=%d", res->ai_addr->sa_family, res->ai_addrlen, conErr, errno);
#endif
      }
#endif
      if (!conErr)
      {
        break;
      }
#ifdef ENABLE_DEBUG
      if (debug)
      {
        printf("didn't find partner\n");
      }
#endif
#ifdef ENABLE_DEBUG2SYSLOG
      if (debug2syslog)
      {
        syslog(LOG_ERR, "NTPclient.setup_receive_generic: didn't find partner");
      }
#endif
      res=res->ai_next;
    }

    if ((!conErr) || (++counter > BACK_OFF_COUNTER_MAX))
    {
      break;
    }
    else
    {   // MR H95366 try more times with delay 1 - 2 - 4 - 8 - 16 - 32 - 64 - 128 secs
      freeaddrinfo(ressave);

      sleep(delay_in_sec);
      delay_in_sec *= 2;
    }
  }

  freeaddrinfo(ressave);
  return conErr;                                  // Attention // new return value // Attention // old:  fd;
}


#if IPv4_ONLY
/* original IPv4 only implementation */
static void setup_transmit(int usd, char *host, short port, struct ntp_control *ntpc)
{
  struct sockaddr_in sa_dest;
  memset(&sa_dest,0,sizeof sa_dest);
  sa_dest.sin_family=AF_INET;
  stuff_net_addr(&(sa_dest.sin_addr),host);
  memcpy(ntpc->serv_addr,&(sa_dest.sin_addr),4); /* XXX asumes IPv4 */
  sa_dest.sin_port=htons(port);
  if (connect(usd,(struct sockaddr *)&sa_dest,sizeof sa_dest)==-1)
  {
    perror("connect");
    exit(1);
  }
}
#endif /* IPv4_ONLY */


/* added support for IPv6 - no functionality at all, so remove it */
#if 0
void setup_transmit_generic(int usd, char *hostname, short port, struct ntp_control *ntpc)
{

  if (debug) printf("transmit_generic: Already connected\n");
}
#endif


static void primary_loop(int usd, struct ntp_control *ntpc)
{
  fd_set fds;
  struct sockaddr_storage sa_xmit;
  int i, pack_len, probes_sent, poll_interval, error;
  unsigned int sa_xmit_len;
  struct timeval to;
  struct ntptime udp_arrival_ntp;
  static u32 incoming_word[325];
#define incoming ((char *) incoming_word)
#define sizeof_incoming (sizeof incoming_word)

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("Listening...\n");
  }
#endif

  probes_sent=0;
  sa_xmit_len=sizeof sa_xmit;
  to.tv_sec=0;
  to.tv_usec=0;
  poll_interval=ntpc->timeout;
#ifdef ENABLE_DEBUG2SYSLOG
  int nDEBcounter = 51;
#endif
  for (;;)
  {
    FD_ZERO(&fds);
    FD_SET(usd,&fds);

    if (signalexit)
    {
	break;  // according to MRI14137 if signal comes before select exit the look and end the process
    }
    i=select(usd+1,&fds,NULL,NULL,&to);  /* Wait on read or error */ 
    if (signalexit)		//select was interrupt by signal
    {
        break;	//according to MRI14137 finish the loop and exit the process
    }
    if ((i!=1)||(!FD_ISSET(usd,&fds)))
    {
      if (i<0)
      {
        if (errno != EINTR)
        {
#ifdef ENABLE_STDERR
          perror("select");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
          {
            if (--nDEBcounter > 0)
            {
              syslog(LOG_ERR, "NTPclient.primary_loop: select: errno=%d (%s). Retry (%d)", errno, strerror(errno), nDEBcounter);
              if (nDEBcounter == 1)
                syslog(LOG_ERR, "NTPclient.primary_loop: No More Retry Logged ***");
            }
          }
#endif
        }
        continue;
      }
      if (to.tv_sec == 0)
      {
        if (probes_sent >= ntpc->probe_count &&
            ntpc->probe_count != 0)
          break;
        send_packet(usd,ntpc->time_of_send);
        ++probes_sent;
        if (ntpc->timeout!=0) /* is defined timeout for no reply? */
        {
          if (to.tv_sec < ntpc->cycle_time)
            to.tv_sec=poll_interval*2+(rand() % (ntpc->timeout/2));
          else
            to.tv_sec=ntpc->cycle_time;
        }
        else
          to.tv_sec=ntpc->cycle_time;
        to.tv_usec=0;
      }
      continue;
    }
    pack_len=recvfrom(usd,incoming,sizeof_incoming,0,
                      (struct sockaddr *)&sa_xmit,&sa_xmit_len);
    if (ntpc->timeout!=0) /* reply received - wait for cycle_time */
      to.tv_sec += abs(ntpc->cycle_time-ntpc->timeout);
    error = ntpc->goodness;
    if (pack_len<0)
    {
#ifdef ENABLE_STDERR
      perror("recvfrom");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
      syslog(LOG_DEBUG, "NTPclient.primary_loop: recvfrom: %s.", strerror(errno));
#endif
    }
    else if (pack_len>0 && (unsigned)pack_len<sizeof_incoming)
    {
      get_packet_timestamp(usd, &udp_arrival_ntp);
      /* Outputs/logs package length and IP address */
      check_source(pack_len, &sa_xmit, sa_xmit_len, ntpc);
      /* rfc1305print() calls "set_time()" depending on its sanity checks */
      if (rfc1305print(incoming_word, &udp_arrival_ntp, ntpc, &error)!=0)
      {
        continue;
      }
      /* udp_handle(usd,incoming,pack_len,&sa_xmit,sa_xmit_len); */
    }
    else
    {
      printf("Ooops.  pack_len=%d\n",pack_len);
#ifdef ENABLE_DEBUG2SYSLOG
      syslog(LOG_ERR, "NTPclient.primary_loop: Wrong pack_len=%d",pack_len);
#endif
      fflush(stdout);
    }
    /* best rollover option: specify -g, -s, and -l.
     * simpler rollover option: specify -s and -l, which
     * triggers a magic -c 1 */
    if (( error < ntpc->goodness && ntpc->goodness != 0 ) ||
        (probes_sent >= ntpc->probe_count && ntpc->probe_count != 0))
    {
      ntpc->set_clock = 0;
      if (!ntpc->live)
        break;
    }
  }
#undef incoming
#undef sizeof_incoming
}


#ifdef ENABLE_REPLAY
static void do_replay(void)
{
  char line[100];
  int n, day, freq, absolute;
  float sec, el_time, st_time, disp;
  double skew, errorbar;
  int simulated_freq = 0;
  unsigned int last_fake_time = 0;
  double fake_delta_time = 0.0;

  while (fgets(line,sizeof line,stdin))
  {
    n=sscanf(line,"%d %f %f %f %lf %f %d",
             &day, &sec, &el_time, &st_time, &skew, &disp, &freq);
    if (n==7)
    {
      fputs(line,stdout);
      absolute=day*86400+(int)sec;
      errorbar=el_time+disp;
#ifdef ENABLE_DEBUG
      if (debug)
      {
        printf("contemplate %u %.1f %.1f %d\n",
               absolute,skew,errorbar,freq);
      }
#endif
      if (last_fake_time==0)
      {
        simulated_freq=freq;
      }
      fake_delta_time += (absolute-last_fake_time)*((double)(freq-simulated_freq))/65536;
#ifdef ENABLE_DEBUG
      if (debug)
      {
        printf("fake %f %d \n", fake_delta_time, simulated_freq);
      }
#endif
      skew += fake_delta_time;
      freq = simulated_freq;
      last_fake_time=absolute;
      simulated_freq = contemplate_data(absolute, skew, errorbar, freq);
    }
    else
    {
#ifdef ENABLE_STDERR
      fprintf(stderr,"Replay input error\n");
#endif
#ifdef ENABLE_DEBUG2SYSLOG
      syslog(LOG_ERR, "NTPclient.do_replay: Replay input error. Stopped.");
#endif
      exit(2);
    }
  }
}
#endif


static void usage(char *argv0)
{
#ifdef ENABLE_STDERR
  fprintf(stderr,
          "Usage: %s [-c count]"
#ifdef ENABLE_DEBUG
          " [-d]"
#endif
#ifdef ENABLE_DEBUG2SYSLOG
          " [-L]"
#endif
          " [-f frequency] [-g goodness] -h hostname\n"
          "\t[-i interval] [-l] [-n timeout] [-p port] [-q min_delay]"
#ifdef ENABLE_REPLAY
          " [-r]"
#endif
          " [-s] [-t]\n",
          argv0);
#else /* ENABLE_STDERR */
  (void) argv0; /* not used */
#endif /* ENABLE_STDERR */
}


int main(int argc, char *argv[])
{
  int usd;  /* socket */
  int c;
  signalexit=0;		    //according to MR I14137 
  /* These parameters are settable from the command line
     the initializations here provide default behavior */
  short int udp_local_port=0;   /* default of 0 means kernel chooses */
  char *hostname=NULL;          /* must be set */
  int initial_freq;             /* initial freq value to use */
  struct ntp_control ntpc;
  ntpc.live=0;
  ntpc.set_clock=0;
  ntpc.probe_count=0;           /* default of 0 means loop forever */
  ntpc.cycle_time=600;          /* seconds */
  ntpc.goodness=0;
  ntpc.cross_check=1;
  ntpc.timeout=60;              /* ignored if 0 */

    
  //according to MR I14137 setup signal handler
  struct sigaction act;
  memset (&act, '\0', sizeof(act));
  act.sa_handler = &signal_handler;
  sigaction(SIGTERM, &act, NULL);


  for (;;)
  {
    c = getopt( argc, argv, "c:" DEBUG_OPTION DEBUG2SYSLOG_OPTION "f:g:h:i:ln:p:q:" REPLAY_OPTION "st");
    if (c == EOF) break;
    switch (c)
    {
    case 'c':
      ntpc.probe_count = atoi(optarg);
      break;
#ifdef ENABLE_DEBUG
    case 'd':
      ++debug;
      break;
#endif
#ifdef ENABLE_DEBUG2SYSLOG
    case 'L':
      ++debug2syslog;
      break;
#endif
    case 'f':
      initial_freq = atoi(optarg);
#ifdef ENABLE_DEBUG
      if (debug)
      {
        printf("initial frequency %d\n",
               initial_freq);
      }
#endif
      set_freq(initial_freq);
      break;
    case 'g':
      ntpc.goodness = atoi(optarg);
      break;
    case 'h':
      hostname = optarg;
      break;
    case 'i':
      ntpc.cycle_time = atoi(optarg);
      break;
    case 'l':
      (ntpc.live)++;
      break;
    case 'n':
      ntpc.timeout = atoi(optarg);
      break;
    case 'p':
      udp_local_port = atoi(optarg);
      break;
    case 'q':
      min_delay = atof(optarg);
      break;
#ifdef ENABLE_REPLAY
    case 'r':
      do_replay();
      exit(0);
      break;
#endif
    case 's':
      (ntpc.set_clock)++;
      break;

    case 't':
      (ntpc.cross_check)=0;
      break;

    default:
      usage(argv[0]);
      exit(1);
    }
  }

  if (hostname == NULL)
  {
    usage(argv[0]);
    exit(1);
  }

  if (ntpc.set_clock && !ntpc.live && !ntpc.goodness && !ntpc.probe_count)
  {
    ntpc.probe_count = 1;
  }

  /* respect only applicable MUST of RFC-4330 */
  if (ntpc.probe_count != 1 && ntpc.cycle_time < MIN_INTERVAL)
  {
    ntpc.cycle_time=MIN_INTERVAL;
  }

#ifdef ENABLE_DEBUG
  if (debug)
  {
    printf("Configuration:\n"
           "  -c probe_count %d\n"
           "  -d (debug)     %d\n"
           "  -g goodness    %d\n"
           "  -h hostname    %s\n"
           "  -i interval    %d\n"
           "  -l live        %d\n"
           "  -n no reply timeout %d\n"
           "  -p local_port  %d\n"
           "  -q min_delay   %f\n"
           "  -s set_clock   %d\n"
           "  -x cross_check %d\n",
           ntpc.probe_count, debug, ntpc.goodness,
           hostname, ntpc.cycle_time, ntpc.live, ntpc.timeout, udp_local_port, min_delay,
           ntpc.set_clock, ntpc.cross_check );
#ifdef ENABLE_DEBUG2SYSLOG
    printf("  -L (syslog)    %d\n",debug2syslog);
#endif
  }
#endif

  syslog(LOG_NOTICE, "NTPclient started ( %s )", hostname);
  srand((unsigned)(time(0)));  /* needed to seed initial time offset */

  /* Startup sequence */
  if ((usd=stuff_net_addr_generic(NULL, NTP_PORT, AF_UNSPEC, SOCK_DGRAM))!= -1)
  {
    /* original IPv4 implementation
     if ((usd=socket(AF_INET,SOCK_DGRAM,IPPROTO_UDP))==-1)
      {perror ("socket");exit(1);}

     setup_receive(usd, INADDR_ANY, udp_local_port);
    */
    // MR H95366 function returns 0 if connected else do exit ntpclient
    if (!setup_receive_generic(hostname, NTP_PORT, AF_UNSPEC, SOCK_DGRAM, usd))
    {
      // connect successful
      primary_loop(usd, &ntpc);

      close(usd);
    }
  }
  return 0;
}

void signal_handler(int signum)
{
	signalexit=(short) signum;
}
